package org.kantega.jexmec.groovy;

/*
 * Copyright 2011 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import groovy.lang.GroovyClassLoader;
import org.codehaus.groovy.control.CompilationFailedException;
import org.kantega.jexmec.ClassLoaderProvider;

import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.util.*;

/**
 *
 */
public class GroovyClassLoaderProvider implements ClassLoaderProvider {

    private final File groovyScriptDirectory;

    private final Set<String> excludedDirectories = new HashSet<String>(Arrays.asList(".svn"));

    private final Map<String, JexmecGroovyClassLoader> classLoaders = new HashMap<String, JexmecGroovyClassLoader>();
    private final Map<String, Long> failedCompilations = new HashMap<String, Long>();
    private Registry registry;
    private ClassLoader parentClassLoader;

    public GroovyClassLoaderProvider(File groovyScriptDirectory) {
        this.groovyScriptDirectory = groovyScriptDirectory;
    }

    public void start(Registry registry, ClassLoader parentClassLoader) {
        this.registry = registry;
        this.parentClassLoader = parentClassLoader;
        if (groovyScriptDirectory != null && groovyScriptDirectory.exists() && groovyScriptDirectory.isDirectory()) {
            scan();
        }
    }

    public synchronized void scan() {

        List<ClassLoader> removedClassLoaders = new ArrayList<ClassLoader>();
        List<ClassLoader> addedClassLoaders = new ArrayList<ClassLoader>();
        final File[] directories = getDirectories();

        Set<String> currentPaths = new HashSet<String>();

        if (directories != null) {
            for (File directory : directories) {

                final File[] scripts = getScripts(directory);

                for (File script : scripts) {
                    currentPaths.add(script.getAbsolutePath());

                    // We might need to remove an existing class loader first
                    if (classLoaders.containsKey(script.getAbsolutePath()) &&
                            classLoaders.get(script.getAbsolutePath()).isChanged()) {
                        removedClassLoaders.add(classLoaders.get(script.getAbsolutePath()));
                        classLoaders.remove(script.getAbsolutePath());
                    }

                    if (!classLoaders.containsKey(script.getAbsolutePath()) &&
                            (failedCompilations.get(script.getAbsolutePath()) == null || !failedCompilations.get(script.getAbsolutePath()).equals(script.lastModified()))) {
                        JexmecGroovyClassLoader loader = new JexmecGroovyClassLoader(script, parentClassLoader);

                        loader.setLastModified(Math.max(script.lastModified(), loader.getLastModified()));

                        GroovyClassLoader groovyClassLoader = new GroovyClassLoader(loader);
                        try {

                            @SuppressWarnings("unchecked")
                            Class groovyClass = groovyClassLoader.parseClass(script);

                            loader.add(groovyClass);

                            classLoaders.put(script.getAbsolutePath(), loader);
                            addedClassLoaders.add(loader);
                            failedCompilations.remove(script.getAbsolutePath());
                        } catch (IOException e) {
                            parsingFailed(script, e);
                        } catch (CompilationFailedException e) {
                            failedCompilations.put(script.getAbsolutePath(), script.lastModified());
                            compilationFailed(script, e);
                        }

                    }
                }
            }
        }
        Set<String> toRemove = new HashSet<String>(classLoaders.keySet());
        toRemove.removeAll(currentPaths);

        for (String path : toRemove) {
            removedClassLoaders.add(classLoaders.get(path));
            classLoaders.remove(path);
        }

        if (addedClassLoaders.size() > 0 || removedClassLoaders.size() > 0) {
            registry.replace(removedClassLoaders, addedClassLoaders);
        }
    }

    protected void parsingFailed(File script, IOException e) {

    }

    protected void compilationFailed(File script, CompilationFailedException e) {

    }

    private File[] getDirectories() {
        return groovyScriptDirectory.listFiles(new FileFilter() {
            public boolean accept(File file) {
                return file.isDirectory() && !excludedDirectories.contains(file.getName());
            }
        });
    }

    private File[] getScripts(File directory) {
        return directory.listFiles(new FileFilter() {
            public boolean accept(File file) {
                return file.isFile() && file.getName().toLowerCase().endsWith(".groovy");
            }
        });
    }

    public void stop() {
        registry.remove(classLoaders.values());
    }

}
